import typing
import statement
import expression
import re
import sys

# This module is responsible for emitting c code for the program,
# using arrays to represent the sets. Specifically, we use an array
# the size the domain of the set, and use a unique identifier for a
# value or tuple to look up if it is in the set.

# Global. Ugly, but it's this or pass around p and out all the time.
p = None
out = None 

# This causes the emission of code that counts how many times things
# have been added to various sets.
set_size_profiling = True

# This enables the h3 hashing model, in which we have exactly one hash
# per h3 set.
# enable_h3 = "h3" in sys.argv

# Disables the print statement. Only profiling output will be printed.
disable_print = False

# Types that keep a count of the net number of allocations.
countable_types = []

##############################################################################
## Utility functions that are used to emit C code.

depth = 0

tempserial = 0
labelserial = 0

# Returns the name of a temporary variable.
def temporary():
    global tempserial
    tempserial += 1

    return "temp%03d" % tempserial

# Returns the name of a label.
def label():
    global labelserial
    labelserial += 1

    return "label%03d" % labelserial

# If c is true, returns a, else b. Not a true conditional: Both a and b
# are evaluated.
def cond(c, a, b):
    if c:
        return a
    else:
        return b


def code(*args):
    out.write("    " * depth + "".join(args) + "\n")

def indent():
    global depth
    depth += 1

def undent():
    global depth
    depth -= 1

def idsub(subs, s):
    """Takes code as input, and replaces identifiers with textual substitutions.

    s -- a string.
    subs -- a map from an identifier to the text it should be substituted with.
    """

    def replace(m):
        if m.group(1) in subs:
            return subs[m.group(1)]
        else:
            return m.group(1)

    return re.sub(r"([a-zA-Z0-9]+)", replace, s)


# This is the nice, easy to use, multi-line coding index. The logic is:
#
# The first argument is a set of substitutions.
#
# The rest of the arguments are iterated through. If an argument is a
# boolean (True or False), it controls which of the next arguments are
# processed.
#
# Otherwised, the line is checked for identifiers [a-zA-Z0-9]+. If an
# identifier is found that is in subs, it is replaced with the
# corresponding text from subs.
def mlcode(subs, *s):
    prefix = True

    for l in s:
        if l is True or l is False:
            prefix = l
            continue
        
        if prefix:
            code(idsub(subs, l))

        prefix = True



##############################################################################
## Information about known expressions at points in the program.

# We only want to add the same thing to a set once. So, when we add an
# expression to a set, or bind it with a for, if, or while statement, we
# check here first to see if the expression has already been bound to a
# variable. If so, returns the variable, otherwise, binds it and
# returns it.

class known_expression:
    """ Represents a known expression that's been added to a variable.

    expr -- the expression in question.
    name -- the name of the variable the expression is bound to.
    """

    def __init__(self, **dict):
        self.__dict__ = dict

# The list of known expressions. This can contain known_expression
# objects, or None. None is used to indicate contexts that can be
# pushed and popped.
kelist = []


def push_ke_context():
    kelist.append(None)

def pop_ke_context():
    while kelist.pop():
        pass

def expression_var(expr):
    for ke in kelist:
        if not ke:
            continue

        if expr == ke.expr and expr.type == ke.expr.type:
            return True, ke.name

    ke = known_expression(expr=expr, name=temporary())
    kelist.append(ke)

    return False, ke.name

##############################################################################
## Data about based sets.

# Constants. It's annoying that python makes me do this this way,
# by maing accessing class data too damn hard.

WEAK = 0
SOME_STRONG = 1
ALL_STRONG = 2



# This is the data we keep about each basing of each set.
class BSD(object):
    def __init__(self, set, strength, impl):

        # Store the implementation. (NA or Hashing)
        self.impl = impl
    
        # Gives a string of w and s, as given by the components of the
        # strength tuple being weak or strong.
        def weak_strong(strength):
            rv = ""

            for i in strength:
                if i:
                    rv += "s"
                else:
                    rv += "w"

            return rv

        # The name the set is associated with.
        self.set = set
        self.name = set + "_" + weak_strong(strength)

        # The strength of each of the elements in the set.
        self.strength = strength

        # The type of the contents of the set.
        self.type = p.tymap[set][0].contains

        # If the contents of this set is a tuple, this is a list
        # of the contents of that tuple. Otherwise, this is equal
        # to [ self.type ].
        self.types = self.type.type_tuple(p.tymap)

        # We have three possible modes that we can be in.
        # WEAK is a weakly-based set.
        # SOME_STRONG is strongly based on at least one but not all elements.
        # ALL_STRONG is strongly based on all elements of the tuple.
        self.mode = WEAK

        if True in strength:
            self.mode = SOME_STRONG

        if False not in strength:
            self.mode = ALL_STRONG

        # This contains the indices of the members of the tuple that are
        # strongly based, but not the primary thing this set is being based
        # off of.
        self.secondary_indices = []
        
        # If we have something declared as an smap, we force it to be
        # strong even if the last element has not been analyzed to be
        # strong.
        if False not in self.strength[:-1] and self.set in p.smaps:
            self.mode = ALL_STRONG
            print "Enabled magic for " + self.name

        # This is true if this set needs to support a remove operation,
        # and false otherwise. We can be more efficient if we do not
        # need to support remove.
        self.support_remove = (self.set in p.needs_remove)

    def emit_definition(self, t):
        # Emit the code to link members of the list together. We only
        # need a list if the set is in some way weak.
        if t is self.type:
            if self.mode != ALL_STRONG:
                code(self.type.c_name + " *" + self.name + "_next;")
            if self.support_remove:
                code(self.type.c_name + " **" + self.name + "_prevnext;")

    # This is responsible for emitting the code to add new entries
    # into the set.
    def emit_add_stmt(self, stmt):
        if stmt.set.variable() != self.set:
            return

        code("// Add to ", self.name)

        known, var = expression_var(stmt.expr)

        if not known:
            code(stmt.expr.type.c_name + " *" +
                 var + " = " + ecode(stmt.expr) +";")


        subs = {
            'name': self.name,
            'var': var,
            }

        # If weak, simply add us to a list.

        if self.mode == WEAK:
            mlcode(subs,
                   self.support_remove, 
                   "if (name) name->name_prevnext = &(var->name_next);",
                   "var->name_next = name;",
                   self.support_remove,
                   "var->name_prevnext = &(name);",
                   "name = var;",
                   )
            return

        # Otherwise, we need to determine what we're strong on.
        strong_ecode = [ e for e, s in zip(ecode_tuple(stmt.expr), self.strength)
                         if s ]

        subs['lcode'] = self.lcode('lookup_create', strong_ecode)
        subs['temp'] = temporary()
        subs['type'] = self.type.c_name

        # If we're somewhat strong, we have to both add to a list and
        # a map. We need to introduce a temporary to store the double
        # pointer to the head of the list, and then we use that to
        # add to the map. This way, we only need to do a single
        # lookup_create.

        if self.mode == SOME_STRONG:

            mlcode(subs,
                   "type **temp = lcode;",
                   self.support_remove,
                   "if (*temp) (*temp)->name_prevnext = (&var->name_next);",
                   "var->name_next = *temp;",
                   self.support_remove,
                   "var->name_prevnext = temp;", 
                   "*temp = var;")
            return

        if self.mode == ALL_STRONG:
            mlcode(subs,
                   "type **temp = lcode;",
                   self.support_remove,
                   "var->name_prevnext = temp;",
                   "*temp = var;")
            return

    def emit_remove_stmt(self, stmt):


        if stmt.set.variable() != self.set:
            return

        assert self.support_remove

        code("// Remove from ", self.name)

        known, var = expression_var(stmt.expr)

        # If this isn't known, then we haven't bothered checking to
        # see if this tuple is in the set.
        assert known

        subs = {
            'name': self.name,
            'var': var,
            }

        if self.mode == ALL_STRONG:
            mlcode(subs,
                   "*(var->name_prevnext) = NULL;")
        else:
            mlcode(subs,
                   "if (var->name_next) var->name_next->name_prevnext = var->name_prevnext;",
                   "*(var->name_prevnext) = var->name_next;")

    def emit_prevnext_ptr_declaration(self):
        if self.mode != WEAK:
            code("%s** %s::*%s_prevnext_ptr = NULL;" %
                 (self.type.c_name, self.type.c_name, self.name))

    def emit_prevnext_ptr_definition(self):
        if self.mode != WEAK and self.support_remove:
            code("%s_prevnext_ptr = &%s::%s_prevnext;" %
                 (self.name, self.type.c_name, self.name))


# This is the data we keep about sets that are based in an RTS manner,
# off of a single member of the tuple.
class RtsBSD(BSD):
 
    def __init__(self, set, strength, impl):
        super(RtsBSD, self).__init__(set, strength, impl)

        # This is the primary type of this based set. The primary type
        # is (arbitrarily) selected as the type of the first member of
        # the tuple that is strongly based.
        self.primary_type = None
        self.primary_index = -1

        for i, s in enumerate(self.strength):

            t = self.types[i]

            if s:
                if self.primary_type == None:
                    self.primary_type = t
                    self.primary_index = i
                else:
                    self.secondary_indices.append(i)
    
    # This is responsible for emitting the definition of this set.
    # When t is a type, emit the aspect of the definition associated
    # with that type. When it's None, emit the global part of the
    # definition.
    def emit_definition(self, t):

        # This emits the global part of the definition.
        super(RtsBSD, self).emit_definition(t)
        
        # WEAK emits a list.
        if self.mode == WEAK:
            if not t:
                code(self.type.c_name + " *" + self.name + ";")
                return
            else:
                return

        # Otherwise, we have a strongly based set. This is first based
        # on the first strong element of the tuple, and then mapped
        # using the second and later strongly based elements.

        # This may generate a map from nothing to a value.
        # This is okay, and we can simply implement a very degenerate
        # map using templates to handle this.

        if self.primary_type != t:
            return

        types = [ self.types[i].c_name for i in self.secondary_indices ]
        types.append(self.type.c_name)
            
        code("%sMap%d<" % (self.impl, len(self.secondary_indices))
             + ", ".join(types) + ", &" + self.name + "_prevnext_ptr> " + self.name + ";")

    def lcode(self, function, args):

        return "%s->%s.%s(%s)" % (
            args[0], self.name, function, ", ".join(args[1:]) )
        

# This handles sets that are represented using the h3 method. (One hash
# per set.)
class H3BSD(BSD):
 
    def __init__(self, set, strength, impl):
        super(H3BSD, self).__init__(set, strength, impl)

    
    # This is responsible for emitting the definition of this set.
    # When t is a type, emit the aspect of the definition associated
    # with that type. When it's None, emit the global part of the
    # definition.
    def emit_definition(self, t):

        # This emits the global part of the definition.
        super(H3BSD, self).emit_definition(t)
        
        # WEAK emits a list.
        if self.mode == WEAK:
            if not t:
                code(self.type.c_name + " *" + self.name + ";")
                return
            else:
                return


        if t:
            return

        strong_types = [ t.c_name for t, s in zip(self.types, self.strength) if s ]

        code("%sMap%d<%s, %s, &%s_prevnext_ptr> %s;"
             % ( self.impl, len(strong_types), ", ".join(strong_types),
                 self.type.c_name, self.name, self.name))
        

    def lcode(self, function, args):
        return "%s.%s(%s)" % (self.name, function, ", ".join(args) )
        

# A list of based_set_data objects.
bsd = []

# This finds out information about all the different possible basings
# of sets in the program.
def build_bsd():

    # We'll always use Rts for a weak set, as these sets are handled
    # by the BSD common code, and it doesn't matter how it's handled.
    none_strong = (RtsBSD, 'NA')

    choices = {
        'rts-na' : (RtsBSD, 'NA'),
        'rts-hashing' : (RtsBSD, 'Hashing'),
        'uniform-na' : (H3BSD, 'NA'),
        'uniform-hashing' : (H3BSD, 'Hashing'),
        }

    one_strong = choices[sys.argv[3]]
    many_strong = choices[sys.argv[4]]
    
    for set in p.bases:
        for strength in p.bases[set]:

            nt = len([x for x in strength if x])

            if nt == 0:
                ctor, impl = none_strong
            elif nt == 1:
                ctor, impl = one_strong
            else:
                ctor, impl = many_strong

            b = ctor(set, strength, impl)

            bsd.append(b)

#     for b in bsd:
#         b.debug_print()

def lookup_bsd(set, strength):

    for i in bsd:
        if i.set == set and i.strength == strength:
            return i

    assert 0


# def gather_types(t, types):
#    if t in types:
#         return

#     types.append(t)   

type_serial = 0

def assign_type_name(t):

    if hasattr(t, "c_name"):
        return

    global type_serial
    type_serial += 1

    t.c_name = "Type%02d" % type_serial

    
##############################################################################
## Type definition code.

def define_strong_sets(t):
    for b in bsd:
        b.emit_definition(t)

def declare_int_type(t):
    code("struct %s;" % t.c_name)

def define_int_type(t):

    subs = {
        'name': t.c_name,
        }

    countable_types.append(t.c_name)

    mlcode(subs,
           "int name_serial = 0;",
           "int name_count = 0;",
           "map<Data *, name *> name_dmap;"
           "",
           "struct name : public TypeBase {")
    
    indent()

    mlcode(subs,
           "Data *data;",
           "",
           "name(Data *d) {",
           "    id = name_serial++;",
           "    data = d;",
           "    name_count++;"
           "}",
           "",
           "~name() {",
           "    name_count--;",
           "}",
           "",
           "static name *discriminate(Data *d) {",
           "    name *rv = name_dmap[d];",
           "    if (! rv) {",
           "        rv = new name(d);",
           "        name_dmap[d] = rv;",
           "        rv->data = d;",
           "    }",
           "    return rv;",
           "}",
           ""
           )

    define_strong_sets(t)

    undent()
    code("};")
    

def define_function_type(t):
    code("// a function.")
    pass

def declare_set_type(t):
    code("struct %s;" % t.c_name)

# I don't believe we will ever use a set directly. We will only access the
# members through the based representations. 
def define_set_type(t):    
    code("// set of " + t.contains.c_name)
    code("struct %s {};" % t.c_name)
    pass

def declare_tuple_type(t):
    code("struct %s;" % t.c_name)

def define_tuple_type(t):

    subs = {
        'name': t.c_name,
        }

    countable_types.append(t.c_name)

    mlcode(subs,
           "map<Data *, name *> name_dmap;",
           "int name_count = 0;",
           "int name_serial = 0; // Never Used.", 
           "",
           "struct name {",
           )
    
    indent()

    # Fields for tuple membership.
    for i, m in enumerate(t.members):
        code("%s *_%d;" % (m.c_name, i))

    # Constructor.
    code()
    code("%s(%s) {" % ( t.c_name,
                        ", ".join(["%s *a_%d" % (m.c_name, i) for i, m in enumerate(t.members) ]) ))
    indent()
    for i, m in enumerate(t.members):
        code("_%d = a_%d;" % (i, i))

    mlcode(subs, "name_count++;")

    undent()
    code("}")
    code()
    
    mlcode(subs,
           "",
           "name() { name_count++; }",
           ""
           "~name() { name_count--; }", 
           "",
           "static name *discriminate(Data *d) {",
           "    name *rv = name_dmap[d];",
           "    if (! rv) {",
           "        rv = new name();",
           "        name_dmap[d] = rv;",
           "    }",
           "    return rv;",
           "}",
           ""
           )

    define_strong_sets(t)

    undent()
    code("};")
    pass

# We emit type definitions by dispatching to an appropriate method on
# this module.
def type_definition(t):
    code("// vars: " + (", ".join([v for v in p.tymap if v and p.tymap[v][0] is t])))

    method = "define_" + type(t).__name__
    globals()[method](t)

def type_declaration(t):
    method = "declare_" + type(t).__name__
    if method in globals():
        globals()[method](t)
    

##############################################################################
def function_definition(name):
    type = p.tymap[name][0]

    code()

    code("Data *%s(%s);" % (name, ", ".join([ "Data *" ] * len(type.args))) )

    code()

#    print type.result.c_name
#    print [str(a) for a in type.args ]

    subs = {
        'type': type.result.c_name,
        'name': name,
        'args': ", ".join([ t.c_name + " *a_" + str(i)
                            for i, t in enumerate(type.args) ]),
        'call': ", ".join([ "a_" + str(i) + "->data" for i, t in enumerate(type.args) ]),
        }

    mlcode(subs,
           "type *name_wrapped(args) {",
           "    Data *rv = name(call);",
           "    return type::discriminate(rv);",
           "}")
        

##############################################################################
# Declarations of variables (inside the method).

def defvar_int_type(v, t):
    code(t.c_name + " *" + v + ";")

def defvar_tuple_type(v, t):
    code(t.c_name + " *" + v + ";")

def defvar(v):

    t = p.tymap[v][0]
    
    method = "defvar_" + t.__class__.__name__
    
    if method in globals():
        globals()[method](v, t)



##############################################################################
## bcode - Evaluate expression in a boolean context.

def bcode(e):
    method = "bcode_" + e.__class__.__name__
    if method in globals():
        return globals()[method](e)
    return ecode(e) + "->data != Zero"

def bcode_and_expr(e):
    return "(" + bcode(e.left) + ") && (" + bcode(e.right) + ")"

def bcode_or_expr(e):
    return "(" + bcode(e.left) + ") || (" + bcode(e.right) + ")"

def bcode_not_expr(e):
    return "!(" + bcode(e.expr) + ")"

# An in that's not used in a for, while, or if must be all strong.
#
# This needs to return 0 or 1, discriminated to the appropriate
# type.
def bcode_in_expr(e, prefix=""):

    b = lookup_bsd(e.set.variable(), e.strength)
    t = ecode_tuple(e.expr)

    # We should use icode when we're dealing with a variable that
    # isn't ALL_STRONG.

    assert b.mode == ALL_STRONG

    check = "*(%s)" % b.lcode('lookup', t)

    # We only need to discriminate if we assign the result somewhere.
    # This is also the only case where we'll have a type with a c_name.
    # if hasattr(e.type, "c_name"):
    #     return prefix + "%s::discriminate((%s) ? %s : %s)" % (e.type.c_name, check, true_, false_)
    # else:

    return prefix + check
    
# We pass to the above, with different options that invert the sense of the
# result.
def bcode_notin_expr(e):
    return bcode_in_expr(e, "!")


##############################################################################
## ecode - Responsible for returning code to evaluate expressions.

# This checks the type of e. If it's not a tuple type, returns
# [ ecode(e) ]. Otherwise, it tries to get a list of expressions
# that are used to construct e. If known, returns the list of those
# expressions, ecoded. Otherwise, returns a list of field acessors
# for e.
        
def ecode_tuple(e):
    if not isinstance(e.type, typing.tuple_type):
        return [ ecode(e) ]

    et = e.tuple()

    if isinstance(et, list):
        return [ ecode(i) for i in et ]

    return [ "(" + ecode(e) + ")->_" + str(i)
             for i in range(0, len(e.type.members)) ]

# Given the expression e, returns the code that's used to evaluate the
# expression in the expression context.
def ecode(e):
    method = "ecode_" + e.__class__.__name__
    if method in globals():
        return globals()[method](e)

    return "%s::discriminate((%s) ? One : Zero)" % (e.type.c_name, bcode(e))

def ecode_tuple_expr(e):
    return "new " + e.type.c_name + "(" + ", ".join([ecode(ex) for ex in e.exprs]) + ")"

# We probably don't need an extra set of parentheses, but they don't
# hurt anything, either.
def ecode_paren_expr(e):
    return "(" + ecode(e.expr) + ")"

def ecode_variable_expr(e):
    return e.name

def ecode_call_expr(e):
    return e.name + "_wrapped(" + ", ".join([ecode(a) for a in e.args]) + ")"

def ecode_integer_expr(e):
    if e.value == 0:
        body = "Zero"
    elif e.value == 1:
        body = "One"
    else:
        body = "IntReading::factory(%d)" % e.value

        
    return e.type.c_name + "::discriminate(" + body + ")"

##############################################################################
## scode - Responsible for emitting code to evaluate statements.

def scode(s):
    method = "scode_" + s.__class__.__name__
    
    if method in globals():
        if type(s) is not list:
            code("// Coding: " + str(s))
        globals()[method](s)
    else:
        code("// %s not found, ignoring." % method)

# This handles a block.
def scode_list(b):
    for s in b:
        scode(s)
        code()

def scode_add_stmt(s):
    if set_size_profiling:
        code("%s_adds++;" % s.set);
        # code('if ((%s_adds & 8191) == 0) fprintf(stderr, "%s %%d\\n", %s_adds);' % (s.set, s.set, s.set))

    for b in bsd:
        b.emit_add_stmt(s)

def scode_remove_stmt(s):
    for b in bsd:
        b.emit_remove_stmt(s)

def scode_program(s):
    scode(s.block)

def scode_for_stmt(s):
    icode(s.expr, lambda : scode(s.block))

# In a while loop, we keep taking the first element of a list. While
# loops therefore only make sense if we _always_ pop off the top
# element of the list. Otherwise, it will never terminate!
def scode_while_stmt(s):

    l = label()
    code(l + ":")
    code()

    def inside():
        scode(s.block)
        code("goto " + l + ";")

    icode(s.expr, inside)

def scode_if_stmt(s):
    l = label()

    def inside():
        push_ke_context()
        scode(s.block)
        pop_ke_context()
        
        code("goto " + l + ";")

    icode(s.expr, inside)

    if s.else_block:
        code()
        code("{")
        indent()
        push_ke_context()
        scode(s.else_block)
        pop_ke_context()
        undent()
        code("}")

    code()
    code(l + ": 0;")

# I should probably upgrade this to handle the assignment of a tuple
# to a tuple. But, why bother?
def scode_assign_stmt(s):
    code(s.left.variable() + " = " + ecode(s.right) + ";")

def scode_abort_stmt(s):
    code('printf("# ABORT: %s\\n");' % s.msg)
    code('return -1;')

def scode_debug_stmt(s):
    if disable_print:
        return

    code('printf("# DEBUG: %s --- ");' % s.msg)
    scode_print_stmt(s)

def scode_profile_stmt(s):
    code('show_profile("%s");' % s.msg)

    # for t in countable_types:
    #     code('printf("# %s_count %%d\\n", %s_count);' % (t, t))
    #     code('printf("# %s_serial %%d\\n", %s_count);' % (t, t))
        
    

def scode_delete_stmt(s):
    code('delete %s;' % ecode(s.expr))

def scode_read_stmt(s):
    exprs = s.expr.force_tuple()

    code("{")
    indent()
    code("Reading **reading = read_readings(%d);" % (len(exprs),) )


    for i, v in enumerate(exprs):
        code()

        t = p.tymap[v.variable()][0]

        # This is really ugly.
        if isinstance(t, typing.set_type):
            read_set(v, t, "reading[%d]" % i)
        else:
            code("%s = %s;" % (v.variable(), rcode(t, "reading[%d]" % i)))

    undent()
    code("}")

# Ugly hack for reading in sets. We fake up an add statement, and then
# emit the code for that. Hey, I did say this was ugly.
def read_set(set, t, base):

    t1 = temporary()
    t2 = temporary()
    iterator = temporary()

    subs = {
        'base': base,
        't1': t1,
        'iii': iterator,
        'ct': t.contains.c_name,
        't2': t2,
        }


    mlcode(subs,
           "SetReading *t1 = dynamic_cast<SetReading *>(base);",
           "for (int iii = 0; iii < t1->contents.size(); iii++) {"
           )

    indent()

    subs['contents'] = rcode(t.contains, "%s->contents[%s]" % (t1, iterator))
    
    mlcode(subs,
           "ct *t2 = contents;")
    
    # At this point, t2 is a variable that points to an object of type ct.
    # Construct and emit an add statement.

    stmt = statement.add_stmt(set=set,
                              expr=expression.variable_expr(name=t2))

    stmt.expr.type = t.contains

    scode(stmt)

    undent()
    code("}")

def scode_print_stmt(s):

    if disable_print:
        return

    exprs = s.expr.force_tuple()

    for e in exprs:
        pcode(e.type, ecode(e))
        code ("printf(\", \");")
        code()

    code ("printf(\"\\n\");")


##############################################################################
## Pcode

# This code is responsible for printing out expressions of the given type,
# from the given variable.

def pcode(t, base):
    method = "pcode_" + type(t).__name__
    return globals()[method](t, base)

def pcode_set_type(t, base):
    tt = t.contains.type_tuple(p.tymap)
    strength = (False,) * len(tt)

    bsd = lookup_bsd(base, strength)

    subs = {
        'name': bsd.name,
        't': temporary(),
        'ctype': t.contains.c_name,
        }

    mlcode(subs,
           'printf("{\\n");',
           "ctype *t = name;",
           "while (t) {",
           )

    indent()
    pcode(t.contains, subs['t'])
    undent()

    mlcode(subs,
           "    ",
           "    printf(\",\\n\");",
           "    t = t->name_next;",
           "}",
           "printf(\"}\");"
           )
    
def pcode_int_type(t, base):
    code(base + "->data->print();")

# This hasn't been tested, and therefore has no known bugs. :-)
def pcode_tuple_type(t, base):

    code('printf("(");')

    for i, ty in enumerate(t.members):
        if i:
            code('printf(", ");')

        pcode(ty, base + "->_%d" % i)

    code('printf(")");')


##############################################################################
## Rcode

# Rcode is used to read in data from reading objects, which are read
# from stdin.

# In these methods, t is a type and base is an expression that evaluates
# to a Reading object, which is where the data is gotten from.

def rcode(t, base):
    method = "rcode_" + type(t).__name__
    return globals()[method](t, base)

def rcode_int_type(t, base):
    return "%s::discriminate(%s)" % (t.c_name, base)

def rcode_tuple_type(t, base):
    temp = temporary()

    code("%s *%s = %s::discriminate(%s);" %
         (t.c_name, temp, t.c_name, base))

    for i, ty in enumerate(t.members):
        code("%s->_%d = %s;" % ( temp, i, 
             rcode(ty, "dynamic_cast<TupleReading *>(%s)->contents[%d]" % (base, i))))

    code()
    
    return temp
    
##############################################################################
## Icode

# This is used to implement the expressions for iteration. Some expressions
# have different semantics when they're used as part of an iteration
# statement. For, while, and if statements are implemented in terms of
# the icode for expressions.

# Icode takes two arguments. The first is an expression, and the second
# is what should go in the center of the code generated for that expression.

# For example, an in statement generates a while loop that iterates over
# the contents of the set.

# inside is a function that takes no arguments that, when called, generates
# the code for the chewey center of the loop. 

def icode(e, inside):
    method = "icode_" + type(e).__name__
    
    if method in globals():
        globals()[method](e, inside)
    else:
        icode_unknown(e, inside)

def icode_paren_expr(e, inside):
    icode(e.expr, inside)

def icode_and_expr(e, inside):
    icode(e.left, lambda : icode(e.right, inside))

# Code for in statements in an iteration context.
def icode_in_expr(e, inside):

    b = lookup_bsd(e.set.variable(), e.strength) 
    strong_ecode = [ ec for ec, st in zip(ecode_tuple(e.expr), e.strength) if st ]

    code("// Iterating over " + e.set.variable())
        
    push_ke_context()

    seen, var = expression_var(e.expr)
    if seen:
        var = temporary()
    
    
    subs = {
        'name': b.name,
        'type': b.type.c_name,
        'var': var
        }

    if b.mode == ALL_STRONG or b.mode == SOME_STRONG:

        subs['lcode'] = b.lcode('lookup', strong_ecode)

        # Do a lookup when we have a strong or partially strong set.        
        mlcode(subs,
               "type *var = *(lcode);")
    else:
        # Otherwise, just grab the first entry in the list.
        mlcode(subs,
               "type *var = name;")

    mlcode(subs,
           "while (var) {")

    indent()

    # Now, we need to assign values to any weakly-bound variables in
    # the expression.

    t = e.expr.tuple()
    if type(t) == list:
        for i, tv in enumerate(t):
            if not e.strength[i]:
                code("%s = %s->_%d;" % (tv.variable(), var, i))
    else:
        if False in e.strength:
            code(t.variable() + " = " + var + ";")

    code()

    # Emit the code for the inside of this iteration.
    inside()

    ## If we're not ALL_STRONG, we have to iterate through all elements
    ## of the list. (This code may be killed if we're in a while or
    ## if loop. Let's hear it for compiler optimization.)

    if b.mode != ALL_STRONG:
        mlcode(subs, "var = var->name_next;")
    else:
        mlcode(subs, "break;")

    pop_ke_context()
    undent()
    code("}")

# For all other expressions, we icode an if statement, and take the
# boolean value.
def icode_unknown(e, inside):
    code("if (" + bcode(e) + ") {")
    indent()
    inside()
    undent()
    code("}")
    

##############################################################################
## Main entry point.

def emit(program, output):
    global p
    p = program

    global out
    out = output

#     print "---> Assigning Names to Types <---"
#    types = []
#    p.walk_types(gather_types, types)

    types = p.find_all_types()
    
    for t in types:
        assign_type_name(t)
    
#     print "---> Building BSD <---"    
    build_bsd()

#     print "---> Emitting Type Definition Code <---"

    code('#include "new_pyapts.h"')
    code('#include <ext/hash_map>')
    code('#include <map>')
    code()
    code('using namespace std;')
    code('using namespace __gnu_cxx;')
    code()
    code("void init_support_code(int, char**);")
    code()

    if set_size_profiling:
        for set in p.bases:
            code('int %s_adds = 0;' % set)            
        code()


    for t in types:
        type_declaration(t)

    code()

    
    code("// prevnext_ptrs:")
    for b in bsd:
        b.emit_prevnext_ptr_declaration()

    code()
      
    for t in types:
        type_definition(t)
        code()

    for b in bsd:
        b.emit_definition(None)
        
    code()

    for i in p.tymap:
        t = p.tymap[i][0]
        if isinstance(t, typing.function_type):
            function_definition(i)

    code()

    code("int main(int argc, char **argv) {")
    indent()

    code()
    code("Data *One = IntReading::factory(1);")
    code("Data *Zero = IntReading::factory(0);")
    code()

    
    for b in bsd:
        b.emit_prevnext_ptr_definition()

    code()
    code("init_support_code(argc, argv);")
    code()

    for v in p.tymap:
        defvar(v)

    code()
    scode(p)

    if set_size_profiling:
        for set in p.bases:
            code('printf("# %s_adds %%d\\n", %s_adds);' % (set, set))

    undent()
    code("}")

import os

if __name__ == "__main__":
    os.system("python main.py reachable.txt")
