import syntax
from typing import *

##############################################################################

class expression(syntax.syntax):

    def __str__(self):
        return self.__class__.format % self.__dict__

    def tuple(self):
        return self

    def force_tuple(self):
        e = self.tuple()
        if isinstance(e, list):
            return e
        return [ e ]

    def variable(self):
        return None

    # We only compare the parts of expressions defined in __fields__ on
    # the class of the expression.
    def __eq__(a, b):

        if type(a) != type(b):
            return False

        for f in type(a).fields:
            if not getattr(a, f) == getattr(b, f):
                return False

        return True

##############################################################################

class and_expr(expression):

    fields = [
        'left',
        'right',
        ]

    format = '%(left)s and %(right)s'

    def find_type(self, imap):
        self.type = int_type()

##############################################################################

class or_expr(expression):

    fields = [
        'left',
        'right',
        ]

    format = '%(left)s or %(right)s'

    def find_type(self, imap):
        self.type = int_type()

##############################################################################

class not_expr(expression):

    fields = [
        'expr',
        ]

    format = 'not %(expr)s'

    def find_type(self, imap):
        self.type = int_type()

##############################################################################

class in_expr(expression):

    fields = [
        'expr',
        'set',
        ]

    format = '%(expr)s in %(set)s'

    def find_type(self, tmap):
        tmap.put(self.set.variable(), set_type(contains=self.expr.type))
        self.type = int_type()

    # This tries to find bindings for variables.
    def bindings(self, p):

        t = self.expr.tuple()
        sv = self.set.variable()

        if sv not in p.bound:
            p.bound.append(sv)

        # We can either have a tuple or a single variable. If we have a single
        # variable, we need to look at the variable.
        if not isinstance(t, list):
            v = t.variable()
            
            if v not in p.bound:
                p.bound.append(v)
                strong = False
            else:
                strong = True                

            # We look up the size of the set, and it's as strong as
            # the size of the set.
            strength = (strong,) * p.tymap[sv][0].contains.len()


        else:
            # Otherwise, we have a tuple. So, compute the strength
            # of it one element at a time.

            strength = []
        
            for e in t:
                v = e.variable()
                if v not in p.bound:
                    strength.append(False)
                    p.bound.append(v)
                else:
                    strength.append(True)
                    

            strength = tuple(strength)
            
        # Now, record what we have.
        self.strength = strength

        bases = p.bases.setdefault(sv, [])
        if strength not in bases:
            bases.append(strength)

        self.set.bindings(p)
        self.expr.bindings(p)





##############################################################################

class notin_expr(expression):

    fields = [
        'expr',
        'set',
        ]

    format = '%(expr)s not in %(set)s'

    def find_type(self, tmap):
        tmap.put(self.set.variable(), set_type(contains=self.expr.type))
        self.type = int_type()


    def bindings(self, p):

        t = self.expr.tuple()
        sv = self.set.variable()

        # Sets are always bound.
        if sv not in p.bound:
            p.bound.append(sv)

        self.expr.bindings(p)
        self.set.bindings(p)

        if not isinstance(t, list):
            v = t.variable()
            strength = (True,) * p.tymap[sv][0].contains.len()
        else:
            strength = (True,) * len(t)
            
        # Now, record what we have.
        self.strength = strength

        bases = p.bases.setdefault(sv, [])

        if strength not in bases:
            bases.append(strength)

##############################################################################

# class binary_expr(expression):
#     fields = [
#         'left',
#         'op',
#         'right'
#         ]

#     format = '%(left)s %(op)s %(right)s'

#     def find_type(self, imap):
#         self.type = int_type()


##############################################################################

class tuple_expr(expression):

    fields = [
        'exprs',
        ]

    def __str__(self):
        return ', '.join(map(str, self.exprs))

    def tuple(self):
        return self.exprs

    def find_type(self, tmap):
        self.type = tuple_type( members = [e.type for e in self.exprs] )


    
##############################################################################

class integer_expr(expression):

    fields = [
        'value',
        ]

    format = '%(value)d'

    def find_type(self, tmap):
        self.type = int_type()
    
##############################################################################

class paren_expr(expression):

    fields = [
        'expr',
        ]

    format = '(%(expr)s)'

    def tuple(self):
        return self.expr.tuple()

    def variable(self):
        return self.expr.variable()

    def find_type(self, tmap):
        self.type = self.expr.type

##############################################################################

class call_expr(expression):

    fields = [
        'name',
        'args',
        ]

    def __str__(self):
        return self.name + "(" + ', '.join(map(str, self.args)) + ")"

    def find_type(self, tmap):
        self.type = int_type()
        tmap.put(self.name,
                 function_type(result = self.type,
                               args = [i.type for i in self.args] ))


##############################################################################

class variable_expr(expression):

    fields = [
        'name',
        ]

    format = '%(name)s'

    def variable(self):
        return self.name

    def find_type(self, tmap):
        self.type = variable_type(name=self.name)
        tmap.setdefault(self.name, [])

    def bindings(self, p):
        if self.name not in p.bound:
            print "Use of unbound variable:", self.name
        

##############################################################################

class subst_expr(expression):

    fields = [
        'name',
        'depth',
        'set'
        ]

    def __str__(self):
        return "%" * self.depth + self.name
        
    def match(self, s, subst):

        if not self.depth == 1:
            return NoMatch, "Depth too high."

        if not self.name in subst:
            subst[self.name] = s
        else:
            # We need tp check and see if we match, but
            # to do that we need to define __eq__ on expressions.
            pass

        return subst

    def duplicate(self, subst):

        if self.depth > 1:
            return subst_expr(name=self.name,
                              depth=self.depth - 1)

        if self.name in subst:
            return subst[self.name].duplicate({})
        
            
