#include <stdio.h>
#include <iostream>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <map>

#include "reading.h"

const int max_treedepth = 10;

using namespace std;

// The purpose of this is to perform the task of reading
// input. This is divided into two sub-tasks. We first read
// in a tree of data. We then turn it into a DAG.

//////////////////////////////////////////////////////////////////////////////


vector<Reading *> all_readings;

Reading::Reading() {
    all_readings.push_back(this);
}

Reading *Reading::replacement() {
    return this;
}

//////////////////////////////////////////////////////////////////////////////

static map<int, IntReading *> irmap;

IntReading::IntReading(int v) {
    value = v;
}

void IntReading::print() {
    printf("%d", value);
}

IntReading *IntReading::factory(int n) {
    IntReading *rv = irmap[n];

    if (! rv) {
        rv = new IntReading(n);
        irmap[n] = rv;
    }

    return rv;
}

int IntReading::leafdist() {
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

static vector<TupleReading *> trlist[max_treedepth];

TupleReading::TupleReading() {    
    _replacement = this;
}

Reading *TupleReading::replacement() {
    return _replacement;
}

void TupleReading::print() {
    printf("(");

    for (int i = 0; i < contents.size(); i++) {
        if (i) printf(", ");
        contents[i]->print();
    }

    printf(")");
}


int TupleReading::leafdist() {
    int max = 0;

    for (int i = 0; i < contents.size(); i++) {
        int d = contents[i]->leafdist();
        if (d > max) max = d;
    }

    trlist[max + 1].push_back(this);

    return max + 1;
}

//////////////////////////////////////////////////////////////////////////////

map<string, StringReading *> srmap;

StringReading *StringReading::factory(char *s) {
    StringReading *rv = srmap[s];

    if (! rv) {
        rv = new StringReading(s);
        srmap[s] = rv;
    }

    return rv;
}

StringReading::StringReading(char *s) : value(s) {
}

void StringReading::print() {
    printf("%s", value.c_str());
}

int StringReading::leafdist() {
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

static vector<FunctorReading *> frlist[max_treedepth];

FunctorReading::FunctorReading(StringReading *name, TupleReading *tuple) {
    this->name = name;
    this->tuple = tuple;
    this->_replacement = this;
}

void FunctorReading::print() {
    name->print();
    tuple->print();
}

Reading *FunctorReading::replacement() {
    return _replacement;
}

int FunctorReading::leafdist() {
    // We know the name is at leafdist 0.
    int d = tuple->leafdist() + 1;

    frlist[d].push_back(this);
    return d;
}

//////////////////////////////////////////////////////////////////////////////

static vector<SetReading *> srlist[max_treedepth];

void SetReading::print() {
    printf("{\n");

    for (int i = 0; i < contents.size(); i++) {
        if (i) printf(",\n");
        contents[i]->print();
    }

    printf("\n}\n");
}

int SetReading::leafdist() {
    int max = 0;

    for (int i = 0; i < contents.size(); i++) {
        int d = contents[i]->leafdist();
        if (d > max) max = d;
    }

    srlist[max + 1].push_back(this);

    return max + 1;
}

//////////////////////////////////////////////////////////////////////////////

// Grammar:
// E = int | tuple | name tuple? | '{' E* '}'
// tuple = '(' E* ')'
//
// There's an ambiguity between name tuple and tuple. We resolve
// this by being greedy.

// Get a character from stdin, or return 0 if we've hit EOF.
char readch() {
    char c;
    
    if (! cin.get(c)) return 0;

    if (c == '#') {

        while (cin.get(c)) {
            if (c == '\n')
                break;
        }

        return readch();
    }

    return c;
}

void unread() {
    cin.unget();
}

Reading *read_E();

TupleReading *read_tuple() {
    TupleReading *rv = new TupleReading();

    while (true) {
        char c = readch();
        while (isspace(c) || c == ',' || c == ':' || c == ';')
            c = readch();

        if (c == ')' || c == ']')
            return rv;

        unread();

        rv->contents.push_back(read_E());
    }
}

Reading *read_E() {

    char buf[1024];
    char *b;
    
    char c = readch();
    while (isspace(c) || c == ',')
        c = readch();

    b = buf;
    
    if (isdigit(c)) {

        while (isdigit(c)) {
            *b++ = c;
            c = readch();
        }

        *b++ = 0;
        unread();

        return IntReading::factory(atoi(buf));        
    }

    if (isalpha(c) || c == '_') {

        while (! isspace(c) && c != ',' && c != '(' && c != ')') {
            *b++ = c;
            c = readch();
        }

        *b++ = 0;

        StringReading *nr = StringReading::factory(buf);
        
        while (isspace(c))
            c = readch();
        
        if (c == '(') {
            return new FunctorReading(nr, read_tuple());
        } else {
            unread();
            return nr;
        }
                    
    }

    if (c == '(' || c == '[') {
        return read_tuple();        
    }

    if (c == '{') {
        SetReading *rv = new SetReading();
        
        while(true) {
            
            c = readch();
            while (isspace(c) || c == ',')
                c = readch();
            
            if (c == '}') {
                return rv;
            }

            unread();

            rv->contents.push_back(read_E());
        }

    }
    
    fprintf(stderr, "I don't understand: %c (%x)\n", c, c);
    abort();
}

//////////////////////////////////////////////////////////////////////////////

void discriminate_tuples(vector<TupleReading *> &tl, int o) {

    // If we have only 1 member, then we're done.
    if (tl.size() <= 1) {
        return;
    }

    map<Reading *, vector<TupleReading *> > split;

    // Otherwise, place the tuples into a map, keyed into the oth
    // element. If there is no oth element, put it in under NULL.
    for (vector<TupleReading *>::iterator i = tl.begin();
         i != tl.end(); i++) {

        TupleReading *tr = *i;
        Reading *r;

        if (tr->contents.size() <= o) {
            r = NULL; 
        } else {
            r = tr->contents[o];
        }

        split[r].push_back(tr);
    }

    // If we haven't been able to discriminate the tuples before hitting
    // the end of the tuple, they must be indistinguishable.
    for (int i = 1; i < split[NULL].size(); i++) {
//        printf("Joining %p to %p.\n", split[NULL][i], split[NULL][0]);
        split[NULL][i]->_replacement = split[NULL][0];
    }

    
    // Otherwise, recurse.
    for (map<Reading *, vector<TupleReading *> >::iterator i = split.begin();
         i != split.end(); i++) {

        if (i->first == NULL) {
            continue;
        }

        discriminate_tuples(i->second, o + 1);        
    }
    
}

// Master discrimination code.
static void discriminate() {

    // For depths moving away from the leavs of the tree.
    for (int d = 0; d < max_treedepth; d++) {

        // For all tuplereadings.
        for (int i = 0; i < trlist[d].size(); i++) {
            TupleReading *tr = trlist[d][i];

            // If a tuple points to something that has already been
            // replaced, handle that replacement.
            for (int j = 0; j < tr->contents.size(); j++) {
                tr->contents[j] = tr->contents[j]->replacement();
            }            
        }

        // Recursively handle tuplereadings.
        discriminate_tuples(trlist[d], 0);
        
        // For all setreadings.
        for (int i = 0; i < srlist[d].size(); i++) {
            SetReading *sr = srlist[d][i];

            // Handle replacements in sets.
            for (int j = 0; j < sr->contents.size(); j++) {
                sr->contents[j] = sr->contents[j]->replacement();
            }
        }

        // We don't bother with discriminating setreadings. If we had
        // to, we could sort and discriminate a set, but that's not
        // immediately necessary, so we don't bother.

        map<Reading *, map <Reading *, FunctorReading *> > frmap;
        
        // For all FunctorReadings...
        for (int i = 0; i < frlist[d].size(); i++) {
            FunctorReading *fr = frlist[d][i];

            fr->tuple = (TupleReading *) fr->tuple->replacement();

            if (frmap[fr->name][fr->tuple]) {
                fr->_replacement = frmap[fr->name][fr->tuple];
            } else {
                frmap[fr->name][fr->tuple] = fr;
            }            
        }        
    }
}

static void cleanup() {
    // Okay, we're done doing the discrimination thing. Let's blow
    // this thing and go home. (Okay, actually we just deallocate
    // useless objects and generally clean up.)
    for (int d = 0; d < max_treedepth; d++) {
        trlist[d].clear();
        frlist[d].clear();
        srlist[d].clear();
    }
        
    // Go through the list of all readings, and get rid of those
    // that have a replacement (that isn't themselves).
    for (int i = 0; i < all_readings.size(); i++) {
        Reading *r = all_readings[i];

        if (r != r->replacement()) {
            delete r;
        }
    }

    all_readings.clear();
}

Reading **read_readings(int n) {
    Reading **rv = new Reading*[n];

    for (int i = 0; i < n; i++) {
        rv[i] = read_E();
        rv[i]->leafdist();
    }

    discriminate();

    for (int i = 0; i < n; i++) {
        rv[i] = rv[i]->replacement();
    }

    cleanup();

    return rv;
    
}

#if 0
int main(int argc, char **argv) {
    Reading *r = read_E();

    r->leafdist();
    discriminate();
    r->print();
}
#endif
