from syntax import *
from typing import *

##############################################################################

class statement(syntax):

    def __init__(self, **args):
        syntax.__init__(self, **args)
        self.replaced = [ ]

    def __str__(self):
        return self.__class__.format % self.__dict__

    def write(self, out, indent):
        out.write(indent)
        out.write(str(self))
        out.write("\n")

    def replace(self, replacement):
        pass


##############################################################################

class compound(statement):

    def write(self, out, indent):
        statement.write(self, out, indent)
        for s in self.block:
            s.write(out, indent + "    ")

            
    def replace(self, r):
        
        for s in self.block:
            if s.replace(r):
                return True

        rv = False
        block = []
            
        for s in self.block:

            if r in s.replaced:
                continue

            s.replaced.append(r)

            try:
                subst = {}
                r.pattern.match(s, subst)

                subst['@'] = s

                for t in r.replacement:
                    block.append(t.duplicate(subst))
                
                rv = True

            except NoMatch:
                block.append(s)


        self.block = block
        
        return rv

##############################################################################

class replace_stmt(statement):

    fields = [
        'pattern',
        'replacement'
        ]

    format = "<replace>"

    def write(self, out, indent):
        out.write(indent)
        out.write("replace:\n")
        self.pattern.write(out, indent + "    ")
        out.write(indent)
        out.write("with:\n")
        for s in self.replacement:
            s.write(out, indent + "    ")


##############################################################################

class while_stmt(compound):

    fields = [
        'expr',
        'block',
        ]

    format = 'while %(expr)s:'

    def bindings(self, p):
        b0 = p.bound[:]
        self.expr.bindings(p)

        for s in self.block:
            s.bindings(p)

        p.bound = b0

##############################################################################

class for_stmt(compound):

    fields = [
        'expr',
        'block',
        ]

    format = 'for %(expr)s:'

    def bindings(self, p):
        b0 = p.bound[:]
        self.expr.bindings(p)

        for s in self.block:
            s.bindings(p)

        p.bound = b0
        

##############################################################################

class if_stmt(compound):

    fields = [
        'expr',
        'block',
        'else_block',
        ]

    format = 'if %(expr)s:'

    def bindings(self, p):
        b0 = p.bound[:]
        self.expr.bindings(p)
        b1 = p.bound[:]

        for s in self.block:
            s.bindings(p)

        p.bound = b1

        for s in self.else_block:
            s.bindings(p)
            
        p.bound = b0

    
##############################################################################

class assign_stmt(statement):

    fields = [
        'left',
        'right',
        ]

    format = '%(left)s = %(right)s'

    def find_type(self, tmap):
        self.left.type = self.right.type
        tmap.put(self.left.variable(),
                 self.right.type)
        pass

    # TODO: Bind the stuff on the left side of this. Extend it to
    # allow a tuple.

    def bindings(self, p):
        v = self.left.variable()

        if v not in p.bound:
            p.bound.append(v)

        self.right.bindings(p)
    
    
##############################################################################

class add_stmt(statement):

    fields = [
        'expr',
        'set',
        ]

    format = '%(set)s add %(expr)s'

    def find_type(self, tmap):
        tmap.put(self.set.variable(),
                 set_type( contains = self.expr.type))

    def bindings(self, p):
        v = self.set.variable()

        if v not in p.bound:
            p.bound.append(v)

        self.expr.bindings(p)

##############################################################################

class remove_stmt(statement):

    fields = [
        'expr',
        'set',
        ]

    format = '%(set)s remove %(expr)s'

    def find_type(self, tmap):
        tmap.put(self.set.variable(),
                 set_type(contains = self.expr.type))

    def bindings(self, p):
        v = self.set.variable()

        if v not in p.needs_remove:
            p.needs_remove.append(v)

        if v not in p.bound:
            p.bound.append(v)

        self.expr.bindings(p)

##############################################################################

class subst_stmt(statement):

    fields = [
        'depth',
        ]

    def __str__(self):
        return "@" * self.depth

    def duplicate(self, subst):
        if self.depth == 1:
            return subst['@']

        return subst_stmt(depth=self.depth-1)

##############################################################################

class invariant_stmt(statement):

    fields = [ 'left',
               'right',
               ]

    format = 'invariant %(left)s = %(right)s'

##############################################################################

class abort_stmt(statement):
    fields = [ 'msg',
               ]

    format = 'abort %(msg)s'

class profile_stmt(statement):
    fields = [ 'msg',
               ]

    format = 'abort %(msg)s'

class debug_stmt(statement):
    fields = [ 'msg',
               'expr',
               ]

    format = 'debug %(msg)s %(expr)s'

    def bindings(self, p):
        exprs = self.expr.force_tuple()

        for e in exprs:
            if isinstance(e.type, set_type):
                l = p.tymap[e.variable()][0].contains.len()
                bases = p.bases.setdefault(e.variable(), [])
                strength = (False,) * l
                if strength not in bases:
                    bases.append(strength)


class read_stmt(statement):
    fields = [ 'expr',
               ]

    format = 'read %(expr)s'

    # Stuff we read in is bound.
    def bindings(self, p):
        t = self.expr.force_tuple()
        for i in t:
            p.bound.append(i.variable())
    
class print_stmt(statement):
    fields = [ 'expr',
               ]

    format = 'print %(expr)s'
    

    # A set that we're printing needs to be weak.
    def bindings(self, p):
        exprs = self.expr.force_tuple()

        for e in exprs:
            e.bindings(p)

            if isinstance(e.type, set_type):
                l = p.tymap[e.variable()][0].contains.len()
                bases = p.bases.setdefault(e.variable(), [])
                strength = (False,) * l
                if strength not in bases:
                    bases.append(strength)

class delete_stmt(statement):
    fields = [ 'expr' ]
    format = 'delete %(expr)s'

class smap_stmt(statement):
    fields = [ 'expr',
               ]

    format = 'smap %(expr)s'

    def bindings(self, p):
        exprs = self.expr.force_tuple()

        for e in exprs:
            if e.variable() not in p.smaps:
                p.smaps.append(e.variable())
        
